@@ -12,14 +12,20 @@ def part_to_child(stu_part, sol_part, append_message, state, node_name=None):
 
     # if the parts are dictionaries, use to deck out child state
     if all(isinstance(p, dict) for p in [stu_part, sol_part]):
-        return state.to_child_state(stu_part['node'], sol_part['node'],
-                                    stu_part.get('target_vars'), sol_part.get('target_vars'),
-                                    stu_part, sol_part,
+        return state.to_child_state(student_subtree=stu_part['node'],
+                                    solution_subtree=sol_part['node'],
+                                    student_context=stu_part.get('target_vars'),
+                                    solution_context=sol_part.get('target_vars'),
+                                    student_parts=stu_part,
+                                    solution_parts=sol_part,
                                     highlight = stu_part.get('highlight'),
-                                    append_message = append_message, node_name=node_name)
+                                    append_message = append_message,
+                                    node_name=node_name)
 
     # otherwise, assume they are just nodes
-    return state.to_child_state(stu_part, sol_part, append_message = append_message)
+    return state.to_child_state(student_subtree=stu_part,
+                                solution_subtree=sol_part,
+                                append_message = append_message)
 
 
 def check_part(name, part_msg, state=None, missing_msg="Are you sure it's defined?", expand_msg=""):
@@ -62,7 +68,7 @@ def check_part_index(name, index, part_msg,
 
 MSG_MISSING = "FMT:The system wants to check the {typestr} you defined but hasn't found it."
 MSG_PREPEND = "__JINJA__:Check your code in the{{' ' + child['part']+ ' of the' if child['part']}} {{typestr}}. "
-def check_node(name, index, typestr, missing_msg=MSG_MISSING, expand_msg=MSG_PREPEND, state=None):
+def check_node(name, index=0, typestr='{ordinal} node', missing_msg=MSG_MISSING, expand_msg=MSG_PREPEND, state=None):
     rep = Reporter.active_reporter
     stu_out = getattr(state, 'student_'+name)
     sol_out = getattr(state, 'solution_'+name)
@@ -170,7 +176,6 @@ def test_not(*args, msg, state=None):
 
     try: multi(*args, state=state)
     except TestFail as e:
-        rep.failed_test = False          # protect against old behavior
         return state
     
     _msg = state.build_message(msg)
@@ -226,6 +231,9 @@ def override(solution, state=None):
 
 from pythonwhat.tasks import setUpNewEnvInProcess, breakDownNewEnvInProcess
 def with_context(*args, state=None):
+
+    rep = Reporter.active_reporter
+
     # set up context in processes
     solution_res = setUpNewEnvInProcess(process = state.solution_process,
                                         context = state.solution_parts['with_items'])
@@ -235,11 +243,11 @@ def with_context(*args, state=None):
     student_res = setUpNewEnvInProcess(process = state.student_process,
                                        context = state.student_parts['with_items'])
     if isinstance(student_res, AttributeError):
-        rep.do_test(Test(Feedback("In your %s `with` statement, you're not using a correct context manager." % (get_ord(index)), child.highlight)))
+        rep.do_test(Test(Feedback("In your `with` statement, you're not using a correct context manager.", child.highlight)))
 
     if isinstance(student_res, (AssertionError, ValueError, TypeError)):
-        rep.do_test(Test(Feedback("In your %s `with` statement, the number of values in your context manager " + \
-            "doesn't correspond to the number of variables you're trying to assign it to." % (get_ord(index)), child.highlight)))
+        rep.do_test(Test(Feedback("In your `with` statement, the number of values in your context manager "
+                                  "doesn't correspond to the number of variables you're trying to assign it to.", child.highlight)))
 
     # run subtests
     try:
@@ -247,13 +255,12 @@ def with_context(*args, state=None):
     finally:
         # exit context
         if breakDownNewEnvInProcess(process = state.solution_process):
-            raise Exception("error in the solution, closing the %s with fails with: %s" %
-                (get_ord(index), close_solution_context))
+            raise Exception("error in the solution, closing the `with` fails with: %s" % (close_solution_context))
 
         if breakDownNewEnvInProcess(process = state.student_process):
 
-            rep.do_test(Test(Feedback("Your %s `with` statement can not be closed off correctly, you're " + \
-                            "not using the context manager correctly." % (get_ord(index)), state.highlight)),
+            rep.do_test(Test(Feedback("Your `with` statement can not be closed off correctly, you're " + \
+                            "not using the context manager correctly.", state.highlight)),
                         fallback_ast = state.highlight)
     return state
 
@@ -261,7 +268,7 @@ def set_context(*args, state=None, **kwargs):
     """Update context values for student and solution environments.
     
     Note that excess args and unmatched kwargs will be unused in the student environment.
-    If an argument is specified both by name and position args, will use named arg.
+    If an argument is specified both by name and position args, ``set_context()`` will use the named arg.
     """
     stu_crnt = state.student_context.context
     sol_crnt = state.solution_context.context
@@ -284,11 +291,74 @@ def set_context(*args, state=None, **kwargs):
     match_keys = dict(zip(sol_crnt.keys(), stu_crnt.keys()))
     out_stu = upd_stu.update({match_keys[k]: v for k,v in kwargs.items() if k in match_keys})
 
-    return state.to_child_state(student_subtree = None, solution_subtree = None,
-                                student_context = out_stu, solution_context = out_sol)
+    return state.to_child_state(student_context = out_stu, solution_context = out_sol)
 
+def set_env(state = None, **kwargs):
+    """Update/set environemnt variables for student and solution environments.
+
+    When ``has_equal_x()`` is used after this, the variables specified through this function will
+    be available in the student and solution process. Note that you will not see these variables
+    in the student process of the state produced by this function: the values are saved on the state
+    and are only added to the student and solution processes when ``has_equal_ast()`` is called.
+    """
+
+    stu_crnt = state.student_env.context
+    sol_crnt = state.solution_env.context
+
+    stu_new = stu_crnt.update(kwargs)
+    sol_new = sol_crnt.update(kwargs)
+
+    return state.to_child_state(student_env = stu_new, solution_env = sol_new)
 
 def check_args(name, missing_msg='FMT:Are you sure it is defined?', state=None):
+    """Check whether a function argument is specified.
+
+    This function can follow ``check_function()`` in an SCT chain and verifies whether an argument is specified.
+    If you want to go on and check whether the argument was correctly specified, you can can continue chaining with
+    ``has_equal_value()`` (value-based check) or ``has_equal_ast()`` (AST-based check)
+
+    This function can also follow ``check_function_def()`` or ``check_lambda_exp()`` to see if arguments have been
+    specified.
+
+    Args:
+        name (str): the name of the argument for which you want to check it is specified. This can also be
+            a number, in which case it refers to the positional arguments. Named argumetns take precedence.
+        missing_msg (str): If specified, this overrides an automatically generated feedback message in case
+            the student did specify the argument.
+        state (State): State object that is passed from the SCT Chain (don't specify this).
+
+    :Examples:
+
+        Student and solution code::
+
+            import numpy as np
+            arr = np.array([1, 2, 3, 4, 5])
+            np.mean(arr)
+
+        SCT::
+
+            # Verify whether arr was correctly set in np.mean
+            # has_equal_value() checks the value of arr, used to set argument a
+            Ex().check_function('numpy.mean').check_args('a').has_equal_value()
+
+            # Verify whether arr was correctly set in np.mean
+            # has_equal_ast() checks the expression used to set argument a
+            Ex().check_function('numpy.mean').check_args('a').has_equal_ast()
+
+        Student and solution code::
+
+            def my_power(x):
+                print("calculating sqrt...")
+                return(x * x)
+
+        SCT::
+
+            Ex().check_function_def('my_power').multi(
+                check_args('x')  # Argument specified by name, will fail if student used y instead.
+                check_args(0)    # Argument specified by position, still passes if student used y instead.
+            )
+
+    """
     if name in ['*args', '**kwargs']:
         return check_part(name, name, state=state, missing_msg = missing_msg)
     else: 
@@ -359,8 +429,43 @@ def call(args,
          argstr='',
          func=None,
          state=None, **kwargs):
+    """Call function definition so you can compare value/output/error generated.
+
+    This function is chained from ``check_function_def()``.
+
+    Args:
+        args: call string or argument list that specifies how the function definition should be called.
+        test (str): Either 'value', 'output' or 'error', specifying if you want to compare the return value,
+            the output generated or the error generated when calling the function.
+        incorrect_msg (str): If specified, this overrides the automatic feedback message when the
+            comparison (value, output or error) is not correct.
+        error_msg (str): If specified, this overrides the automatic message that is generated when
+            the call generated an error (when it shouldn't) or didn't generate one (when it should).
+        argstr (str): argument for backwards compatibility.
+        func (function): binary function that tells you how the comparison should be.
+        state (State): state object that is chained from.
+
+    :Example:
+
+        Student and solution code::
+
+            def my_power(x):
+                print("calculating sqrt...")
+                return(x * x)
+
+        SCT::
+
+            Ex().check_function_def('my_power').multi(
+                call("my_power(3)", "value"),  # specified as string, compare return value
+                call([3], "value"),            # specifies as list, compare return value
+                call([3], "output")            # specified as list, compare output generated
+            )
+
+    """
+
     rep = Reporter.active_reporter
-    test_type = ('value', 'output', 'error')
+
+    assert test in ('value', 'output', 'error')
 
     get_func = evalCalls[test]
 
@@ -400,6 +505,11 @@ from pythonwhat import utils
 def has_equal_ast(incorrect_msg="FMT: Your code does not seem to match the solution.", code=None, exact=True, state=None):
     """Test whether abstract syntax trees match between the student and solution code.
 
+    ``has_equal_ast()`` can be used in two ways:
+
+    * As a robust version of ``has_code()``. By setting ``code``, you can look for the AST representation of ``code`` in the student's submission.
+    * As an expression-based check when using more advanced SCT chain, e.g. to compare the equality of expressions to set function arguments.
+
     Args:
         incorrect_msg: message displayed when ASTs mismatch.
         code: optional code to use instead of the solution AST
@@ -419,6 +529,20 @@ def has_equal_ast(incorrect_msg="FMT: Your code does not seem to match the solut
             Ex().has_equal_ast(code = "dict(a = 'value').keys()")
             Ex().has_equal_ast(code = "dict(a = 'value')", exact = False)
 
+        Student and Solution Code::
+
+            import numpy as np
+            arr = np.array([1, 2, 3, 4, 5])
+            np.mean(arr)
+
+        SCT::
+
+            # Check underlying value of arugment a of np.mean:
+            Ex().check_function('numpy.mean').check_args('a').has_equal_ast()
+
+            # Only check AST equality of expression used to specify argument a:
+            Ex().check_function('numpy.mean').check_args('a').has_equal_ast()
+
     """
     rep = Reporter.active_reporter
 
@@ -441,51 +565,23 @@ def has_equal_ast(incorrect_msg="FMT: Your code does not seem to match the solut
 
     return state
 
-def has_expr(incorrect_msg="__JINJA__:Unexpected expression {{test}}: expected `{{sol_eval}}`, got `{{stu_eval}}`{{' with values ' + extra_env if extra_env}}.",
-             error_msg="Running an expression in the student process caused an issue.",
-             undefined_msg="FMT:Have you defined `{name}` without errors?",
+DEFAULT_INCORRECT_MSG="__JINJA__:Unexpected expression {{test}}: expected `{{sol_eval}}`, got `{{stu_eval}}`{{' with values ' + extra_env if extra_env}}."
+DEFAULT_ERROR_MSG="Running an expression in the student process caused an issue."
+DEFAULT_UNDEFINED_MSG="FMT:Have you defined `{name}` without errors?"
+def has_expr(incorrect_msg=DEFAULT_INCORRECT_MSG,
+             error_msg=DEFAULT_ERROR_MSG,
+             undefined_msg=DEFAULT_UNDEFINED_MSG,
              extra_env=None,
              context_vals=None,
-             expr_code=None,
              pre_code=None,
-             keep_objs_in_env=None,
+             expr_code=None,
              name=None,
              highlight=None,
              copy=True,
              func=None,
              state=None,
              test=None):
-    """Run student and solution code, compare returned value, printed output, or errors.
 
-    Args:
-        incorrect_msg (str): feedback message if the output of the expression in the solution doesn't match
-          the one of the student. This feedback message will be expanded if it is used in the context of
-          another test function, like test_if_else.
-        error_msg (str): feedback message if there was an error when running the student code.
-          Note that when testing for an error, this message is displayed when none is raised.
-        undefined_msg (str): feedback message if the name argument is defined, but a variable 
-          with that name doesn't exist after running the student code.
-        extra_env (dict): set variables to the extra environment. They will update the student
-          and solution environment in the active state before the student/solution code in the active
-          state is ran. This argument should contain a dictionary with the keys the names of
-          the variables you want to set, and the values are the values of these variables.
-        context_vals (list): set variables which are bound in a for loop to certain values. This argument is
-          only useful if you use the function in a test_for_loop. It contains a list with the values
-          of the bound variables.
-        expr_code (str): if this variable is not None, the expression in the student/solution code will not
-          be ran. Instead, the given piece of code will be ran in the student as well as the solution environment
-          and the result will be compared.
-        pre_code (str): the code in string form that should be executed before the expression is executed.
-          This is the ideal place to set a random seed, for example.
-        keep_obj_in_env (list()): a list of variable names that should be hold in the copied environment where
-          the expression is evaluated. All primitive types are copied automatically, other objects have to
-          be passed explicitely.
-        name (str): the name of a variable, or expression, whose value will be tested after running the
-          student and solution code. This could be thought of as post code.
-        copy (bool): whether to try to deep copy objects in the environment, such as lists, that could
-          accidentally be mutated. Disable to speed up SCTs. Disabling may lead to cryptic mutation issues.
-        func: custom binary function of form f(stu_result, sol_result), for equality testing.
-    """
     rep = Reporter.active_reporter
 
     # run function to highlight a block of code
@@ -499,24 +595,25 @@ def has_expr(incorrect_msg="__JINJA__:Unexpected expression {{test}}: expected `
                        context_vals = context_vals,
                        pre_code = pre_code,
                        expr_code = expr_code,
-                       keep_objs_in_env = keep_objs_in_env,
-                       name=name,
+                       name = name,
                        copy=copy,
                        do_exec = True if test == 'output' else False)
     
     eval_sol, str_sol = get_func(tree = state.solution_tree,
                                  process = state.solution_process,
-                                 context = state.solution_context)
+                                 context = state.solution_context,
+                                 env = state.solution_env)
 
     if (test == 'error') ^ isinstance(str_sol, Exception):
-        raise ValueError("evaluating expression raised error in solution process (or not an error if testing for one). "
-                         "Error: %s - %s"%(type(str_sol), str_sol))
+        raise ValueError("Evaluating expression raised error in solution process (or not an error if testing for one). "
+                         "Error: {} - {}".format(type(str_sol), str_sol))
     if isinstance(eval_sol, ReprFail):
         raise ValueError("Couldn't figure out the value of a default argument: " + eval_sol.info)
 
     eval_stu, str_stu = get_func(tree = state.student_tree,
                                  process = state.student_process,
-                                 context = state.student_context)
+                                 context = state.student_context,
+                                 env = state.student_env)
 
     # kwargs ---
     fmt_kwargs = {'stu_part': state.student_parts, 'sol_part': state.solution_parts, 
@@ -543,6 +640,77 @@ def has_expr(incorrect_msg="__JINJA__:Unexpected expression {{test}}: expected `
 
     return state
 
+
+
+args_string = """
+
+    Args:
+        incorrect_msg (str): feedback message if the {0} of the expression in the solution
+          doesn't match the one of the student. This feedback message will be expanded if it is used
+          in the context of another check function, like ``check_if_else``.
+        error_msg (str): feedback message if there was an error when running the targeted student code.
+          Note that when testing for an error, this message is displayed when none is raised.
+        undefined_msg (str): feedback message if the ``name`` argument is defined, but a variable
+          with that name doesn't exist after running the targeted student code.
+        extra_env (dict): set variables to the extra environment. They will update the student and solution environment in
+          the active state before the student/solution code in the active state is ran. This argument should contain a
+          dictionary with the keys the names of the variables you want to set, and the values are the values of these variables.
+          You can also use ``set_env()`` for this.
+        context_vals (list): set variables which are bound in a ``for`` loop to certain values.
+          This argument is only useful when checking a for loop (or list comprehensions).
+          It contains a list with the values of the bound variables.
+          You can also use ``set_context()`` for this.
+        pre_code (str): the code in string form that should be executed before the expression is executed.
+          This is the ideal place to set a random seed, for example.
+        expr_code (str): if this argument is set, the expression in the student/solution code will not
+          be ran. Instead, the given piece of code will be ran in the student as well as the solution environment
+          and the result will be compared.
+        name (str): If this is specified, the {1} of running this expression after running the focused expression
+          is returned, instead of the {1} of the focussed expression in itself. This is typically used to inspect the
+          {1} of an object after executing the body of e.g. a ``for`` loop.
+        copy (bool): whether to try to deep copy objects in the environment, such as lists, that could
+          accidentally be mutated. Disable to speed up SCTs. Disabling may lead to cryptic mutation issues.
+        func: custom binary function of form f(stu_result, sol_result), for equality testing.
+    """
+
 has_equal_value =  partial(has_expr, test = 'value')
+has_equal_value.__doc__ = """Run targeted student and solution code, and compare returned value.
+
+    When called on an SCT chain, ``has_equal_value()`` will execute the student and solution
+    code that is 'zoomed in on' and compare the returned values.
+    """ + args_string.format("returned value", "value") + """
+    :Example:
+
+        Student code and solution code::
+
+            import numpy as np
+            arr = np.array([1, 2, 3, 4, 5])
+            np.mean(arr)
+
+        SCT::
+
+            # Verify equality of arr:
+            Ex().check_object('arr').has_equal_value()
+
+            # Verify whether arr was correctly set in np.mean
+            Ex().check_function('numpy.mean').check_args('a').has_equal_value()
+
+            # Verify whether np.mean(arr) produced the same result
+            Ex().check_function('numpy.mean').has_equal_value()
+
+    """
+
+
 has_equal_output = partial(has_expr, test = 'output')
+has_equal_output.__doc__ = """Run targeted student and solution code, and compare output.
+
+    When called on an SCT chain, ``has_equal_output()`` will execute the student and solution
+    code that is 'zoomed in on' and compare the output.
+    """ + args_string.format("output", "output")
+
 has_equal_error  = partial(has_expr, test = 'error')
+has_equal_error.__doc__ = """Run targeted student and solution code, and compare generated errors.
+
+    When called on an SCT chain, ``has_equal_error()`` will execute the student and solution
+    code that is 'zoomed in on' and compare the errors that they generate.
+    """ + args_string.format("error", "error")
